/*
 *	DEFS.H
 *	Tom Kerrigan's Simple Chess Program (TSCP)
 *
 *	Copyright 1997 Tom Kerrigan
 */

#define eng_name "TSCP"
#define eng_version "zh 1.1"

#define BOOL			int
#define TRUE			1
#define FALSE			0

#define GEN_STACK		65536 /* zh:, higher than normal chess */
#define MAX_PLY			32
#define HIST_STACK		400

#define LIGHT			0
#define DARK			1

#define PAWN			0
#define KNIGHT			1
#define BISHOP			2
#define ROOK			3
#define QUEEN			4
#define P_KNIGHT        5  /* zh: promoted pawn to knight */
#define P_BISHOP        6  /* zh: */
#define P_ROOK          7  /* zh: */
#define P_QUEEN         8  /* zh: */
#define KING			9  /* zh: */
#define EMPTY			10 /* zh: */

/* useful squares */
#define A1				56
#define B1				57
#define C1				58
#define D1				59
#define E1				60
#define F1				61
#define G1				62
#define H1				63
#define A8				0
#define B8				1
#define C8				2
#define D8				3
#define E8				4
#define F8				5
#define G8				6
#define H8				7

#define ROW(x)			(x >> 3)
#define COL(x)			(x & 7)

/* zh: */
#define ABS(x)          (((x) < 0)? (-(x)) : (x))
#define MAX(a,b)  (((a) > (b)) ? (a) : (b))
#define MIN(a,b)  (((a) < (b)) ? (a) : (b))
#define FILE_DIST(a,b) (ABS((COL(a)) - (COL(b))))
#define RANK_DIST(a,b) (ABS((ROW(a)) - (ROW(b))))
#define DIST(a,b) (MAX(FILE_DIST(a,b), RANK_DIST(a,b)))


/* This is the basic description of a move. promote is what
   piece to promote the pawn to, if the move is a pawn
   promotion. drop is a piece from on-hand for crazyhouse variant,
   example P@g7, the P here is the drop. bits is a bitfield 
   that describes the move, with the following bits:

   1	capture
   2	castle
   4	en passant capture
   8	pushing a pawn 2 squares
   16	pawn move
   32	promote
   64   drop move

   It's union'ed with an integer so two moves can easily
   be compared with each other. */

typedef struct {
	char from;
	char to;
	char promote;
	char bits;
	char drop; /* zh: */
} move_bytes;

typedef union {
	move_bytes b;
	int u;
} move;

/* an element of the move stack. it's just a move with a
   score, so it can be sorted by the search functions. */
typedef struct {
	move m;
	int score;
} gen_t;

/* an element of the history stack, with the information
   necessary to take a move back. */
typedef struct {
	move m;
	int capture;
	int castle;
	int ep;
	int fifty;
	int hash;
} hist_t;
