/*
 *	MAIN.C
 *	Tom Kerrigan's Simple Chess Program (TSCP)
 *
 *	Copyright 2016 Tom Kerrigan
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include "defs.h"
#include "data.h"
#include "protos.h"


/* get_ms() returns the milliseconds elapsed since midnight,
   January 1, 1970. */

#include <sys/timeb.h>
BOOL ftime_ok = FALSE;  /* does ftime return milliseconds? */
int get_ms()
{
	struct timeb timebuffer;
	ftime(&timebuffer);
	if (timebuffer.millitm != 0)
		ftime_ok = TRUE;
	return (timebuffer.time * 1000) + timebuffer.millitm;
}


/* main() is basically an infinite loop that either calls
   think() when it's the computer's turn to move or prompts
   the user for a command (and deciphers it). */

int main()
{
	int computer_side;
	char s[256];
	int m;

	printf("\n");
	printf("Tom Kerrigan's Simple Chess Program (TSCP)\n");
	printf("version 1.81b, 3/10/16\n");
	printf("Copyright 2016 Tom Kerrigan\n");
	printf("\n");
	printf("%s %s\n", eng_name, eng_version);
	printf("A crazyhouse chess variant engine\n");
	printf("\n");
	printf("\"help\" displays a list of commands.\n");
	printf("\n");
	init_hash();
	init_board();
	gen();
	computer_side = EMPTY;
	max_time = 1 << 25;
	max_depth = 4;
	option_use_book = 0;
	for (;;) {
		if (side == computer_side) {  /* computer's turn */
			
			/* think about the move and make it */
			think(1);
			if (!pv[0][0].u) {
				printf("(no legal moves)\n");
				computer_side = EMPTY;
				continue;
			}
			printf("Computer's move: %s\n", move_str(pv[0][0].b));
			makemove(pv[0][0].b);
			ply = 0;
			gen();
			print_result();
			continue;
		}

		/* get user input */
		printf("tscp zh> ");
		if (scanf("%s", s) == EOF)
			return 0;
		if (!strcmp(s, "on")) {
			computer_side = side;
			continue;
		}
		if (!strcmp(s, "off")) {
			computer_side = EMPTY;
			continue;
		}
		if (!strcmp(s, "st")) {
			scanf("%d", &max_time);
			max_time *= 1000;
			max_depth = 32;
			continue;
		}
		if (!strcmp(s, "sd")) {
			scanf("%d", &max_depth);
			max_time = 1 << 25;
			continue;
		}
		if (!strcmp(s, "undo")) {
			if (!hply)
				continue;
			computer_side = EMPTY;
			takeback();
			ply = 0;
			gen();
			continue;
		}
		if (!strcmp(s, "new")) {
			computer_side = EMPTY;
			init_board();
			gen();
			continue;
		}
		if (!strcmp(s, "d")) {
			print_board();
			continue;
		}
		if (!strcmp(s, "bench")) {
			computer_side = EMPTY;
			bench();
			continue;
		}
		if (!strcmp(s, "bye")) {
			printf("Share and enjoy!\n");
			break;
		}
		if (!strcmp(s, "xboard")) {
			xboard();
			break;
		}
		if (!strcmp(s, "help")) {
			printf("on - computer plays for the side to move\n");
			printf("off - computer stops playing\n");
			printf("st n - search for n seconds per move\n");
			printf("sd n - search n ply per move\n");
			printf("undo - takes back a move\n");
			printf("new - starts a new game\n");
			printf("d - display the board\n");
			printf("bench - run the built-in benchmark\n");
			printf("bye - exit the program\n");
			printf("xboard - switch to XBoard mode\n");
			printf("Enter moves in coordinate notation, e.g., e2e4, e7e8Q\n");
			continue;
		}

		/* maybe the user entered a move? */
		m = parse_move(s);
		if (m == -1 || !makemove(gen_dat[m].m.b))
			printf("Illegal move.\n");
		else {
			ply = 0;
			gen();
			print_result();
		}
	}
	if (option_use_book)
		close_book();
	return 0;
}


/* parse the move s (in coordinate notation) and return the move's
   index in gen_dat, or -1 if the move is illegal */

int parse_move(char *s)
{
	int from, to, i;
	BOOL drop_move = FALSE;

	/* zh: drop move test */
	if((s[0]=='P' || s[0]=='N' || s[0]=='B' || s[0]=='R' || s[0]=='Q') && s[1]=='@'){
		if(s[2] < 'a' || s[2] > 'h' || s[3] < '0' || s[3] > '9') return -1;
		drop_move = TRUE;
	}
	else {
		/* make sure the string looks like a move */
		if (s[0] < 'a' || s[0] > 'h' ||
			s[1] < '0' || s[1] > '9' ||
			s[2] < 'a' || s[2] > 'h' ||
			s[3] < '0' || s[3] > '9')
			return -1;
	}

	if(drop_move){
		to = s[2] - 'a';
		to += 8 * (8 - (s[3] - '0'));

		for (i = 0; i < first_move[1]; ++i){
			if(gen_dat[i].m.b.bits&64 && gen_dat[i].m.b.to == to){
				if(gen_dat[i].m.b.drop == PAWN && s[0]=='P'){
					return i;
				}
				if(gen_dat[i].m.b.drop == KNIGHT && s[0]=='N'){
					return i;
				}
				if(gen_dat[i].m.b.drop == BISHOP && s[0]=='B'){
					return i;
				}
				if(gen_dat[i].m.b.drop == ROOK && s[0]=='R'){
					return i;
				}
				if(gen_dat[i].m.b.drop == QUEEN && s[0]=='Q'){
					return i;
				}
			}			
		}
	}
	// Else if not a drop move
	else {
		if (s[0] < 'a' || s[0] > 'h' ||
			s[1] < '0' || s[1] > '9' ||
			s[2] < 'a' || s[2] > 'h' ||
			s[3] < '0' || s[3] > '9')
			return -1;

		from = s[0] - 'a';
		from += 8 * (8 - (s[1] - '0'));
		to = s[2] - 'a';
		to += 8 * (8 - (s[3] - '0'));

		for (i = 0; i < first_move[1]; ++i)
			if (gen_dat[i].m.b.from == from && gen_dat[i].m.b.to == to) {

				/* if the move is a promotion, handle the promotion piece;
				assume that the promotion moves occur consecutively in
				gen_dat. */
				if (gen_dat[i].m.b.bits & 32){
					switch (s[4]) {
					case 'n':
					case 'N':
						return i;
					case 'b':
					case 'B':
						return i + 1;
					case 'r':
					case 'R':
						return i + 2;
					case 'q':
					case 'Q':
						return i + 3;
					default:  /* assume it's a queen */
						return i + 3;
					}
				}
				return i;
			}
	}

	/* didn't find the move */
	return -1;
}


/* move_str returns a string with move m in coordinate notation */

char *move_str(move_bytes m)
{
	static char str[6];

	char c;

	/* zh: construct a drop move text */
	if (m.bits & 64) {
		char drop_pc = m.drop;
		if (drop_pc == PAWN) {
			sprintf(str, "P@%c%d", COL(m.to) + 'a', 8 - ROW(m.to));
			str[4] = 0;
		}
		else if (drop_pc == KNIGHT) {
			sprintf(str, "N@%c%d", COL(m.to) + 'a', 8 - ROW(m.to));
			str[4] = 0;
		}
		else if (drop_pc == BISHOP) {
			sprintf(str, "B@%c%d", COL(m.to) + 'a', 8 - ROW(m.to));
			str[4] = 0;
		}
		else if (drop_pc == ROOK) {
			sprintf(str, "R@%c%d", COL(m.to) + 'a', 8 - ROW(m.to));
			str[4] = 0;
		}
		else if (drop_pc == QUEEN) {
			sprintf(str, "Q@%c%d", COL(m.to) + 'a', 8 - ROW(m.to));
			str[4] = 0;
		}
	}
	else if (m.bits & 32) {
		switch (m.promote) {
		case P_KNIGHT: /* zh: typical promoted piece type */
			c = 'n';
			break;
		case P_BISHOP:
			c = 'b';
			break;
		case P_ROOK:
			c = 'r';
			break;
		case P_QUEEN:
			c = 'q';
			break;
		default:
			c = 'q';
			break;
		}
		sprintf(str, "%c%d%c%d%c",
			COL(m.from) + 'a',
			8 - ROW(m.from),
			COL(m.to) + 'a',
			8 - ROW(m.to),
			c);
	}
	else
		sprintf(str, "%c%d%c%d",
		COL(m.from) + 'a',
		8 - ROW(m.from),
		COL(m.to) + 'a',
		8 - ROW(m.to));

	return str;
}


/* print_board() prints the board */

void print_board()
{
	int i;
	
	printf("\n8 ");
	for (i = 0; i < 64; ++i) {
		switch (color[i]) {
			case EMPTY:
				printf(" .");
				break;
			case LIGHT:
				printf(" %c", piece_char[piece[i]]);
				break;
			case DARK:
				printf(" %c", piece_char[piece[i]] + ('a' - 'A'));
				break;
		}
		if ((i + 1) % 8 == 0 && i != 63)
			printf("\n%d ", 7 - ROW(i));
	}
	printf("\n\n   a b c d e f g h\n\n");

	/* zh: show onhand pieces */
	printf("onhand, white, black\n");
	printf("%6s, %4d  %4d\n", "queen", onhand_piece[LIGHT][QUEEN], onhand_piece[DARK][QUEEN]);
	printf("%6s, %4d  %4d\n", "rook", onhand_piece[LIGHT][ROOK], onhand_piece[DARK][ROOK]);
	printf("%6s, %4d  %4d\n", "bishop", onhand_piece[LIGHT][BISHOP], onhand_piece[DARK][BISHOP]);
	printf("%6s, %4d  %4d\n", "knight", onhand_piece[LIGHT][KNIGHT], onhand_piece[DARK][KNIGHT]);
	printf("%6s, %4d  %4d\n\n", "pawn", onhand_piece[LIGHT][PAWN], onhand_piece[DARK][PAWN]);
}


/* xboard() is a substitute for main() that is XBoard
   and WinBoard compatible. See the following page for details:
   http://www.research.digital.com/SRC/personal/mann/xboard/engine-intf.html */

void xboard()
{
	int computer_side;
	char line[256], command[256];
	int m;
	int post = 0;
	int ping_num = 0; /* zh: */
	base_time = 0; /* zh: init */
	period = 0; /* zh: init */
	inc = 0;    /* zh: init */
	

	signal(SIGINT, SIG_IGN);
	printf("\n");
	init_board();
	gen();
	computer_side = EMPTY;
	for (;;) {
		fflush(stdout);
		if (side == computer_side) {
			think(post);
			if (!pv[0][0].u) {
				computer_side = EMPTY;
				continue;
			}
			printf("move %s\n", move_str(pv[0][0].b));
			makemove(pv[0][0].b);
			ply = 0;
			gen();
			print_result();
			continue;
		}
		if (!fgets(line, 256, stdin))
			return;
		if (line[0] == '\n')
			continue;
		sscanf(line, "%s", command);
		if (!strcmp(command, "xboard"))
			continue;
		if (!strcmp(command, "new")) {
			init_board();
			gen();
			computer_side = DARK;
			continue;
		}
		if (!strcmp(command, "quit"))
			return;
		if (!strcmp(command, "force")) {
			computer_side = EMPTY;
			continue;
		}
		if (!strcmp(command, "white")) {
			side = LIGHT;
			xside = DARK;
			gen();
			computer_side = DARK;
			continue;
		}
		if (!strcmp(command, "black")) {
			side = DARK;
			xside = LIGHT;
			gen();
			computer_side = LIGHT;
			continue;
		}
		if (!strcmp(command, "st")) {
			sscanf(line, "st %d", &max_time);
			max_time *= 1000;
			max_depth = 32;
			continue;
		}
		if (!strcmp(command, "sd")) {
			sscanf(line, "sd %d", &max_depth);
			max_time = 1 << 25;
			continue;
		}
		if (!strcmp(command, "time")) {
			sscanf(line, "time %d", &max_time);
			max_time *= 10;
			max_time /= 30;
			max_depth = 32;
			/* zh: If there is inc add more time to max_time */
			if (inc)
				max_time += 8*inc/10;
			continue;
		}
		/* zh: level <period> <tc> <inc> */
 		if (!strcmp(command, "level")) {
			sscanf(line, "level %d %d %d", &period, &base_time, &inc);
			base_time *= 60*1000; /* zh: Convert base_time from minute to ms */
			max_time = base_time/30; /* zh: time allocated for a single move */
			max_depth = 32;
			inc *= 1000; /* zh: Convert increment from sec to ms */
			continue;
		}
		if (!strcmp(command, "otim")) {
			continue;
		}
		if (!strcmp(command, "go")) {
			computer_side = side;
			continue;
		}
		if (!strcmp(command, "hint")) {
			think(0);
			if (!pv[0][0].u)
				continue;
			printf("Hint: %s\n", move_str(pv[0][0].b));
			continue;
		}
		if (!strcmp(command, "undo")) {
			if (!hply)
				continue;
			takeback();
			ply = 0;
			gen();
			continue;
		}
		if (!strcmp(command, "remove")) {
			if (hply < 2)
				continue;
			takeback();
			takeback();
			ply = 0;
			gen();
			continue;
		}
		if (!strcmp(command, "post")) {
			post = 2;
			continue;
		}
		if (!strcmp(command, "nopost")) {
			post = 0;
			continue;
		}
		/* zh: */
		if (!strcmp(command, "ping")) {
			sscanf(line, "ping %d", &ping_num);
			printf("pong %d\n", ping_num);
			continue;
		}
		/* zh: */
		if (!strcmp(command, "variant")) {
			continue;
		}
		/* zh: Options */
		if (!strcmp(command, "option")) {
			char option_name[128];
			char *value;	
			int v;
			sscanf(line, "option %s", option_name);

			// Book
			value = strstr(option_name, "Book=");
			if (value) {
				value += 5;
				v = atoi(value);
				option_use_book = v;
				if(option_use_book)
					open_book();
				else
					close_book();
				continue;
			}
			continue;
		}
		/* zh: protover command is new in winboard protocol version 2.
		   feature variants=crazyhouse tells GUI that this engine
		   is capable of playing this variant. Also added feature myname
		   to let GUI know the engine name. */
		if (!strcmp (command, "protover")){
			printf("feature time=1 colors=1 ping=1\n");
			printf("feature sigint=0 sigterm=0 reuse=0\n");
			printf("feature variants=\"crazyhouse\"\n");			
			printf("feature myname=\"%s %s\"\n", eng_name, eng_version);
			printf("feature option=\"Book -check %d\"\n", option_use_book);
            printf("feature done=1\n");
			continue;
		}
		m = parse_move(line);
		if (m == -1 || !makemove(gen_dat[m].m.b))
			printf("Error (unknown command): %s\n", command);
		else {
			ply = 0;
			gen();
			print_result();
		}
	}
}


/* print_result() checks to see if the game is over, and if so,
   prints the result. */

void print_result()
{
	int i;

	/* is there a legal move? */
	for (i = 0; i < first_move[1]; ++i)
		if (makemove(gen_dat[i].m.b)) {
			takeback();
			break;
		}
	if (i == first_move[1]) {
		if (in_check(side)) {
			if (side == LIGHT)
				printf("0-1 {Black mates}\n");
			else
				printf("1-0 {White mates}\n");
		}
		else
			printf("1/2-1/2 {Stalemate}\n");
	}
	else if (reps() == 2)
		printf("1/2-1/2 {Draw by repetition}\n");
	else if (fifty >= 100)
		printf("1/2-1/2 {Draw by fifty move rule}\n");
}


/* bench: This is a little benchmark code that calculates how many
   nodes per second TSCP searches.
   This is just the standard starting position of crazyhouse.
   It searches six ply three times. It calculates nodes per
   second from the best time. */

/* zh: start color */
int bench_color[64] = {
	1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1,
	EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY,
	EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY,
	EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY,
	EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0
};

/* zh: start pieces */
int bench_piece[64] = {
	3, 1, 2, 4, KING, 2, 1, 3,
	0, 0, 0, 0, 0, 0, 0, 0,
	EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY,
	EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY,
	EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY,
	EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY,
	0, 0, 0, 0, 0, 0, 0, 0,
	3, 1, 2, 4, KING, 2, 1, 3
};

void bench()
{
	int i;
	int t[3];
	double nps;

	/* setting the position to a non-initial position confuses the opening
	   book code. */
	close_book();

	king_sq[LIGHT] = E1;
	king_sq[DARK] = E8;
	for (i = 0; i < 64; ++i) {
		color[i] = bench_color[i];
		piece[i] = bench_piece[i];
	}
	side = LIGHT;
	xside = DARK;
	castle = 0;
	ep = -1;
	fifty = 0;
	ply = 0;
	hply = 0;
	set_hash();
	print_board();
	max_time = 1 << 25;
	max_depth = 6;
	for (i = 0; i < 3; ++i) {
		think(1);
		t[i] = get_ms() - start_time;
		printf("Time: %d ms\n", t[i]);
	}
	if (t[1] < t[0])
		t[0] = t[1];
	if (t[2] < t[0])
		t[0] = t[2];
	printf("\n");
	printf("Nodes: %d\n", nodes);
	printf("Best time: %d ms\n", t[0]);
	if (!ftime_ok) {
		printf("\n");
		printf("Your compiler's ftime() function is apparently only accurate\n");
		printf("to the second. Please change the get_ms() function in main.c\n");
		printf("to make it more accurate.\n");
		printf("\n");
		return;
	}
	if (t[0] == 0) {
		printf("(invalid)\n");
		return;
	}
	nps = (double)nodes / (double)t[0];
	nps *= 1000.0;

	printf("Nodes per second: %d\n", (int)nps);

	init_board();
	if (option_use_book)
		open_book();
	gen();
}
